""" Permissions for ai-aside API"""

from rest_framework.permissions import BasePermission

from ai_aside.config_api.validators import validate_course_key
from ai_aside.platform_imports import can_change_summaries_settings


class HasStudioWriteAccess(BasePermission):
    """
    Check if the user has studio write access to a course.
    """
    def has_permission(self, request, view):
        """
        Check permissions for this class.
        """

        if not request.user.is_authenticated:
            return False

        if not request.user.is_active:
            return False

        course_key_string = view.kwargs.get('course_id')
        course_key = validate_course_key(course_key_string)

        return can_change_summaries_settings(request.user, course_key)



"""Constants for AI-Aside."""

# Regex for Course ID URL patterns
COURSE_ID_REGEX = r'[^/+]+(/|\+)[^/+]+(/|\+)[^/]+'
COURSE_ID_PATTERN = r'(?P<course_id>{})'.format(COURSE_ID_REGEX)

# Regex for Usage ID URL patterns (Unit IDs)
UNIT_ID_REGEX = r'(?:i4x://?[^/]+/[^/]+/[^/]+/[^@]+(?:@[^/]+)?)|(?:[^/]+)'
UNIT_ID_PATTERN = r'(?P<unit_id>{})'.format(UNIT_ID_REGEX)

"""
Constants used by DjangoXBlockUserService
"""

# Optional attributes stored on the XBlockUser

# The personally identifiable user ID.
ATTR_KEY_USER_ID = 'edx-platform.user_id'
# The user's role in the course ('staff', 'instructor', or 'student').
ATTR_KEY_USER_ROLE = 'edx-platform.user_role'



"""
Config API Utilities
"""
from edx_rest_framework_extensions.auth.jwt.authentication import JwtAuthentication
from edx_rest_framework_extensions.auth.session.authentication import SessionAuthentication
from rest_framework import status
from rest_framework.response import Response
from rest_framework.views import APIView

from ai_aside.config_api.exceptions import AiAsideException
from ai_aside.config_api.permissions import HasStudioWriteAccess


class APIResponse(Response):
    """API Response"""
    def __init__(self, data=None, http_status=None, content_type=None, success=False):
        _status = http_status or status.HTTP_200_OK
        data = data or {}
        reply = {'response': {'success': success}}
        reply['response'].update(data)
        super().__init__(data=reply, status=_status, content_type=content_type)


class AiAsideAPIView(APIView):
    """
    Base API View with authentication and permissions.
    """

    authentication_classes = (JwtAuthentication, SessionAuthentication,)
    permission_classes = (HasStudioWriteAccess,)


def handle_errors(view_func):
    """
    Wrapper which handles our standard exception.

    We cannot do this by overriding handle_exception as you might expect,
    because the newrelic wrapper sits between the view function and the
    handle_exception and logs it, which makes our expected exceptions seem
    harmful. So we'll handle those before newrelic can see them.
    """
    def wrapped_viewfunc(self_, request, **kwargs):
        try:
            return view_func(self_, request, **kwargs)
        except AiAsideException as exc:
            return APIResponse(http_status=exc.http_status, data={'message': str(exc)})
    return wrapped_viewfunc



"""
Utilities related to API views
"""

from opaque_keys import InvalidKeyError
from opaque_keys.edx.keys import CourseKey, UsageKey

from ai_aside.config_api.exceptions import AiAsideException


def validate_course_key(course_key_string: str) -> CourseKey:
    """
    Validate and parse a course_key string, if supported.
    """
    try:
        course_key = CourseKey.from_string(course_key_string)
    except InvalidKeyError as error:
        raise AiAsideException(f"{course_key_string} is not a valid CourseKey") from error
    if course_key.deprecated:
        raise AiAsideException("Deprecated CourseKeys (Org/Course/Run) are not supported.")
    return course_key


def validate_unit_key(unit_key_string: str) -> UsageKey:
    """
    Validate and parse a unit_key string, if supported.
    """
    try:
        usage_key = UsageKey.from_string(unit_key_string)
    except InvalidKeyError as error:
        raise AiAsideException(f"{unit_key_string} is not a valid UsageKey") from error
    return usage_key



"""
Waffle flag controlling summary hook distribution.

Xblocks do not generally use waffle flags,
import at use time method taken from the LTI Xblock
"""

# Namespace
WAFFLE_NAMESPACE = 'summaryhook'


# .. toggle_name: summaryhook.staff_only
# .. toggle_implementation: CourseWaffleFlag
# .. toggle_default: False
# .. toggle_description: Enables OpenAI driven summary xblock aside for staff only
# .. toggle_use_cases: temporary
# .. toggle_creation_date: 2023-03-14
# .. toggle_target_removal_date: 2023-12-06
# .. toggle_tickets: ACADEMIC-15709 (2U)
# .. toggle_warning: None.
SUMMARYHOOK_STAFF_ONLY = 'summaryhook_staff_only'


# .. toggle_name: summaryhook.summaries_configuration
# .. toggle_implementation: CourseWaffleFlag
# .. toggle_default: False
# .. toggle_description: Enables summaries configuration for course and/or unit
# .. toggle_use_cases: temporary
# .. toggle_creation_date: 2023-07-24
# .. toggle_target_removal_date: 2024-04-06
# .. toggle_tickets: ACADEMIC-16209 (2U)
# .. toggle_warning: None.
SUMMARYHOOK_SUMMARIES_CONFIGURATION = 'summaryhook_summaries_configuration'


def _is_summaryhook_waffle_flag_enabled(flag_name, course_key):
    """
    Import and return Waffle flag for enabling the summary hook.
    """
    # pylint: disable=import-outside-toplevel
    try:
        from openedx.core.djangoapps.waffle_utils import CourseWaffleFlag
        return CourseWaffleFlag(f'{WAFFLE_NAMESPACE}.{flag_name}', __name__).is_enabled(course_key)
    except ImportError:
        return False


def summary_staff_only(course_key):
    """
    Return whether the summaryhook.summaryhook_staff_only WaffleFlag is on.
    """
    return _is_summaryhook_waffle_flag_enabled(SUMMARYHOOK_STAFF_ONLY, course_key)


def summaries_configuration_enabled(course_key):
    """
    Return whether the summaryhook.summaryhook_summaries_configuration WaffleFlag is on.
    """
    return _is_summaryhook_waffle_flag_enabled(SUMMARYHOOK_SUMMARIES_CONFIGURATION, course_key)



"""
Implements a simple REST API for updating unit and course settings.

Setters:
    POST: ai_aside/v1/:course_id - (payload: { enabled: True/False })
    POST: ai_aside/v1/:course_id/:unit_id - (payload: { enabled: True/False })

Getters:
    GET: ai_aside/v1/:course_id - (response: { success: True/False, enabled: True/False })
    GET: ai_aside/v1/:course_id/:unit_id - (response: { success: True/False, enabled: True/False })

Delete:
    DELETE: ai_aside/v1/:course_id - (response: { success: True/False })
    DELETE: ai_aside/v1/:course_id/:unit_id - (response: { success: True/False })

Both GET and DELETE methods respond with a 404 if the setting cannot be found.
"""
from ai_aside.config_api.api import (
    delete_course_settings,
    delete_unit_settings,
    get_course_settings,
    get_unit_settings,
    is_summary_config_enabled,
    reset_course_unit_settings,
    set_course_settings,
    set_unit_settings,
)
from ai_aside.config_api.exceptions import AiAsideException, AiAsideNotFoundException
from ai_aside.config_api.validators import validate_course_key, validate_unit_key
from ai_aside.config_api.view_utils import AiAsideAPIView, APIResponse, handle_errors


class CourseSummaryConfigEnabledAPIView(AiAsideAPIView):
    """
    Simple GET endpoint to expose whether the course may use summary config.
    """
    @handle_errors
    def get(self, request, course_id=None):
        """Expose whether the course may use summary config"""
        if course_id is None:
            raise AiAsideNotFoundException

        course_key = validate_course_key(course_id)
        enabled = is_summary_config_enabled(course_key)
        return APIResponse(success=True, data={'enabled': enabled})


class CourseEnabledAPIView(AiAsideAPIView):
    """Handlers for course level settings"""
    @handle_errors
    def get(self, request, course_id=None):
        """Gets the enabled state for a course"""
        if course_id is None:
            raise AiAsideNotFoundException

        course_key = validate_course_key(course_id)
        settings = get_course_settings(course_key)
        return APIResponse(success=True, data=settings)

    @handle_errors
    def post(self, request, course_id=None):
        """Update the course and reset if its necessary"""

        # enabled: Updates the course enabled default state
        enabled = request.data.get('enabled')

        # reset: If it is present, it will delete all unit settings, resetting them back to the default
        reset = request.data.get('reset')

        try:
            course_key = validate_course_key(course_id)
            set_course_settings(course_key, {'enabled': enabled})
            if reset:
                reset_course_unit_settings(course_key)
        except TypeError as error:
            raise AiAsideException('Invalid parameters') from error

        return APIResponse(success=True)

    @handle_errors
    def delete(self, request, course_id=None):
        """Deletes the settings for a module"""

        if course_id is None:
            raise AiAsideNotFoundException

        course_key = validate_course_key(course_id)
        delete_course_settings(course_key)
        return APIResponse(success=True)


class UnitEnabledAPIView(AiAsideAPIView):
    """Handlers for module level settings"""
    @handle_errors
    def get(self, request, course_id=None, unit_id=None):
        """Gets the enabled state for a unit"""
        if course_id is None or unit_id is None:
            raise AiAsideNotFoundException

        course_key = validate_course_key(course_id)
        unit_key = validate_unit_key(unit_id)
        settings = get_unit_settings(course_key, unit_key)
        return APIResponse(success=True, data=settings)

    @handle_errors
    def post(self, request, course_id=None, unit_id=None):
        """Sets the enabled state for a unit"""

        enabled = request.data.get('enabled')

        try:
            course_key = validate_course_key(course_id)
            unit_key = validate_unit_key(unit_id)
            set_unit_settings(course_key, unit_key, {'enabled': enabled})
        except TypeError as error:
            raise AiAsideException('Invalid parameters') from error

        return APIResponse(success=True)

    @handle_errors
    def delete(self, request, course_id=None, unit_id=None):
        """Deletes the settings for a unit"""

        if course_id is None or unit_id is None:
            raise AiAsideNotFoundException

        course_key = validate_course_key(course_id)
        unit_key = validate_unit_key(unit_id)
        delete_unit_settings(course_key, unit_key,)
        return APIResponse(success=True)



"""
Custom exceptions for ai-aside
"""
from rest_framework import status


class AiAsideException(Exception):
    """
    A common base class for all exceptions
    """
    http_status = status.HTTP_400_BAD_REQUEST


class AiAsideNotFoundException(AiAsideException):
    """
    A 404 exception class
    """
    http_status = status.HTTP_404_NOT_FOUND
